@@ -1,6 +1,6 @@
 /*
  * BlueALSA - dbus-client.c
- * Copyright (c) 2016-2019 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -11,12 +11,15 @@
 #include "shared/dbus-client.h"
 
 #include <errno.h>
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
+#include "shared/defs.h"
+
 static int path2ba(const char *path, bdaddr_t *ba) {
 
 	unsigned int x[6];
@@ -67,6 +70,9 @@ dbus_bool_t bluealsa_dbus_connection_ctx_init(
 	if ((ctx->conn = dbus_bus_get_private(DBUS_BUS_SYSTEM, error)) == NULL)
 		return FALSE;
 
+	/* do not terminate in case of D-Bus connection being lost */
+	dbus_connection_set_exit_on_disconnect(ctx->conn, FALSE);
+
 	if (!dbus_connection_set_watch_functions(ctx->conn, ba_dbus_watch_add,
 				ba_dbus_watch_del, ba_dbus_watch_toggled, ctx, NULL)) {
 		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
@@ -154,6 +160,26 @@ dbus_bool_t bluealsa_dbus_connection_signal_match_clean(
 	return TRUE;
 }
 
+/**
+ * Dispatch D-Bus messages synchronously. */
+dbus_bool_t bluealsa_dbus_connection_dispatch(
+		struct ba_dbus_ctx *ctx) {
+
+	struct pollfd fds[8];
+	nfds_t nfds = ARRAYSIZE(fds);
+
+	bluealsa_dbus_connection_poll_fds(ctx, fds, &nfds);
+	if (poll(fds, nfds, 0) > 0)
+		bluealsa_dbus_connection_poll_dispatch(ctx, fds, nfds);
+
+	/* Dispatch incoming D-Bus messages/signals. The actual dispatching is
+	 * done in a function registered with dbus_connection_add_filter(). */
+	while (dbus_connection_dispatch(ctx->conn) == DBUS_DISPATCH_DATA_REMAINS)
+		continue;
+
+	return TRUE;
+}
+
 dbus_bool_t bluealsa_dbus_connection_poll_fds(
 		struct ba_dbus_ctx *ctx,
 		struct pollfd *fds,
@@ -211,6 +237,90 @@ dbus_bool_t bluealsa_dbus_connection_poll_dispatch(
 	return rv;
 }
 
+/**
+ * Callback function for BlueALSA service properties parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
+		DBusMessageIter *value, void *userdata, DBusError *error) {
+	struct ba_service_props *props = (struct ba_service_props *)userdata;
+
+	char type = dbus_message_iter_get_arg_type(value);
+	char type_expected;
+
+	if (strcmp(key, "Version") == 0) {
+		if (type != (type_expected = DBUS_TYPE_STRING))
+			goto fail;
+		const char *tmp;
+		dbus_message_iter_get_basic(value, &tmp);
+		strncpy(props->version, tmp, sizeof(props->version) - 1);
+	}
+	else if (strcmp(key, "Adapters") == 0) {
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+		const char *tmp[ARRAYSIZE(props->adapters)];
+		size_t length = ARRAYSIZE(tmp);
+		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+			return FALSE;
+		if (length > ARRAYSIZE(props->adapters))
+			length = ARRAYSIZE(props->adapters);
+		for (size_t i = 0; i < length; i++)
+			strncpy(props->adapters[i], tmp[i], sizeof(props->adapters[i]) - 1);
+	}
+
+	return TRUE;
+
+fail:
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect variant for '%s': %c != %c", key, type, type_expected);
+	return FALSE;
+}
+
+/**
+ * Get properties of BlueALSA service. */
+dbus_bool_t bluealsa_dbus_get_props(
+		struct ba_dbus_ctx *ctx,
+		struct ba_service_props *props,
+		DBusError *error) {
+
+	static const char *interface = BLUEALSA_INTERFACE_MANAGER;
+	DBusMessage *msg = NULL, *rep = NULL;
+	dbus_bool_t ret = FALSE;
+
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, "/org/bluealsa",
+					DBUS_INTERFACE_PROPERTIES, "GetAll")) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	DBusMessageIter iter;
+	dbus_message_iter_init_append(msg, &iter);
+	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &interface)) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
+					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
+		goto fail;
+
+	if (!dbus_message_iter_init(rep, &iter)) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE, "Empty response message");
+		goto fail;
+	}
+
+	if (!bluealsa_dbus_message_iter_dict(&iter, error,
+				bluealsa_dbus_message_iter_get_props_cb, props))
+		goto fail;
+
+	ret = TRUE;
+
+fail:
+	if (rep != NULL)
+		dbus_message_unref(rep);
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	return ret;
+}
+
 dbus_bool_t bluealsa_dbus_get_pcms(
 		struct ba_dbus_ctx *ctx,
 		struct ba_pcm **pcms,
@@ -226,7 +336,6 @@ dbus_bool_t bluealsa_dbus_get_pcms(
 
 	dbus_bool_t rv = TRUE;
 	struct ba_pcm *_pcms = NULL;
-	char *signature;
 	size_t i;
 
 	DBusMessage *rep;
@@ -245,8 +354,13 @@ dbus_bool_t bluealsa_dbus_get_pcms(
 			dbus_message_iter_get_arg_type(&iter_pcms) != DBUS_TYPE_INVALID;
 			dbus_message_iter_next(&iter_pcms), i++) {
 
-		if (dbus_message_iter_get_arg_type(&iter_pcms) != DBUS_TYPE_DICT_ENTRY)
-			goto fail_signature;
+		if (dbus_message_iter_get_arg_type(&iter_pcms) != DBUS_TYPE_DICT_ENTRY) {
+			char *signature = dbus_message_iter_get_signature(&iter);
+			dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+					"Incorrect signature: %s != a{oa{sv}}", signature);
+			dbus_free(signature);
+			goto fail;
+		}
 
 		struct ba_pcm *tmp = _pcms;
 		if ((tmp = realloc(tmp, (i + 1) * sizeof(*tmp))) == NULL) {
@@ -273,11 +387,6 @@ dbus_bool_t bluealsa_dbus_get_pcms(
 
 	goto success;
 
-fail_signature:
-	signature = dbus_message_iter_get_signature(&iter);
-	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
-			"Incorrect signature: %s != a{oa{sv}}", signature);
-
 fail:
 	if (_pcms != NULL)
 		free(_pcms);
@@ -293,47 +402,59 @@ success:
 dbus_bool_t bluealsa_dbus_get_pcm(
 		struct ba_dbus_ctx *ctx,
 		const bdaddr_t *addr,
-		unsigned int flags,
+		unsigned int transports,
+		unsigned int mode,
 		struct ba_pcm *pcm,
 		DBusError *error) {
 
+	const bool get_last = bacmp(addr, BDADDR_ANY) == 0;
 	struct ba_pcm *pcms = NULL;
+	struct ba_pcm *match = NULL;
 	dbus_bool_t rv = TRUE;
 	size_t length = 0;
+	uint32_t seq = 0;
 	size_t i;
 
 	if (!bluealsa_dbus_get_pcms(ctx, &pcms, &length, error))
 		return FALSE;
 
-	for (i = 0; i < length; i++)
-		if (bacmp(&pcms[i].addr, addr) == 0 &&
-				(pcms[i].flags & flags) == flags) {
-			memcpy(pcm, &pcms[i], sizeof(*pcm));
-			goto final;
+	for (i = 0; i < length; i++) {
+		if (get_last) {
+			if (pcms[i].sequence >= seq &&
+					pcms[i].transport & transports &&
+					pcms[i].mode == mode) {
+				seq = pcms[i].sequence;
+				match = &pcms[i];
+			}
 		}
+		else if (bacmp(&pcms[i].addr, addr) == 0 &&
+				pcms[i].transport & transports &&
+				pcms[i].mode == mode) {
+			match = &pcms[i];
+			break;
+		}
+	}
 
-	rv = FALSE;
-	dbus_set_error(error, DBUS_ERROR_FILE_NOT_FOUND, "PCM not found");
+	if (match != NULL)
+		memcpy(pcm, match, sizeof(*pcm));
+	else {
+		dbus_set_error(error, DBUS_ERROR_FILE_NOT_FOUND, "PCM not found");
+		rv = FALSE;
+	}
 
-final:
 	free(pcms);
 	return rv;
 }
 
-dbus_bool_t bluealsa_dbus_pcm_open(
+/**
+ * Open BlueALSA PCM stream. */
+dbus_bool_t bluealsa_dbus_open_pcm(
 		struct ba_dbus_ctx *ctx,
 		const char *pcm_path,
-		int operation_mode,
 		int *fd_pcm,
 		int *fd_pcm_ctrl,
 		DBusError *error) {
 
-	const char *mode = NULL;
-	if (operation_mode == BA_PCM_FLAG_SOURCE)
-		mode = "source";
-	else if (operation_mode == BA_PCM_FLAG_SINK)
-		mode = "sink";
-
 	DBusMessage *msg;
 	if ((msg = dbus_message_new_method_call(ctx->ba_service, pcm_path,
 					BLUEALSA_INTERFACE_PCM, "Open")) == NULL) {
@@ -341,12 +462,6 @@ dbus_bool_t bluealsa_dbus_pcm_open(
 		return FALSE;
 	}
 
-	if (!dbus_message_append_args(msg, DBUS_TYPE_STRING, &mode, DBUS_TYPE_INVALID)) {
-		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
-		dbus_message_unref(msg);
-		return FALSE;
-	}
-
 	DBusMessage *rep;
 	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
 					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL) {
@@ -365,7 +480,9 @@ dbus_bool_t bluealsa_dbus_pcm_open(
 	return rv;
 }
 
-dbus_bool_t bluealsa_dbus_rfcomm_open(
+/**
+ * Open BlueALSA RFCOMM socket for dispatching AT commands. */
+dbus_bool_t bluealsa_dbus_open_rfcomm(
 		struct ba_dbus_ctx *ctx,
 		const char *rfcomm_path,
 		int *fd_rfcomm,
@@ -395,6 +512,66 @@ dbus_bool_t bluealsa_dbus_rfcomm_open(
 	return rv;
 }
 
+/**
+ * Update BlueALSA PCM property. */
+dbus_bool_t bluealsa_dbus_pcm_update(
+		struct ba_dbus_ctx *ctx,
+		const struct ba_pcm *pcm,
+		enum ba_pcm_property property,
+		DBusError *error) {
+
+	static const char *interface = BLUEALSA_INTERFACE_PCM;
+	const char *_property = NULL;
+	const char *variant = NULL;
+	const void *value = NULL;
+	int type = -1;
+
+	switch (property) {
+	case BLUEALSA_PCM_SOFT_VOLUME:
+		_property = "SoftVolume";
+		variant = DBUS_TYPE_BOOLEAN_AS_STRING;
+		value = &pcm->soft_volume;
+		type = DBUS_TYPE_BOOLEAN;
+		break;
+	case BLUEALSA_PCM_VOLUME:
+		_property = "Volume";
+		variant = DBUS_TYPE_UINT16_AS_STRING;
+		value = &pcm->volume.raw;
+		type = DBUS_TYPE_UINT16;
+		break;
+	}
+
+	DBusMessage *msg;
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, pcm->pcm_path,
+					DBUS_INTERFACE_PROPERTIES, "Set")) == NULL)
+		goto fail;
+
+	DBusMessageIter iter;
+	DBusMessageIter iter_val;
+
+	dbus_message_iter_init_append(msg, &iter);
+	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &interface) ||
+			!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &_property) ||
+			!dbus_message_iter_open_container(&iter, DBUS_TYPE_VARIANT, variant, &iter_val) ||
+			!dbus_message_iter_append_basic(&iter_val, type, value) ||
+			!dbus_message_iter_close_container(&iter, &iter_val))
+		goto fail;
+
+	if (!dbus_connection_send(ctx->conn, msg, NULL))
+		goto fail;
+
+	dbus_message_unref(msg);
+	return TRUE;
+
+fail:
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+	return FALSE;
+}
+
+/**
+ * Send command to the BlueALSA PCM controller socket. */
 dbus_bool_t bluealsa_dbus_pcm_ctrl_send(
 		int fd_pcm_ctrl,
 		const char *command,
@@ -426,6 +603,42 @@ dbus_bool_t bluealsa_dbus_pcm_ctrl_send(
 	return TRUE;
 }
 
+/**
+ * Extract strings from the string array. */
+dbus_bool_t bluealsa_dbus_message_iter_array_get_strings(
+		DBusMessageIter *iter,
+		DBusError *error,
+		const char **strings,
+		size_t *length) {
+
+	char *signature;
+	size_t i;
+
+	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_ARRAY)
+		goto fail;
+
+	DBusMessageIter iter_array;
+	for (dbus_message_iter_recurse(iter, &iter_array), i = 0;
+			dbus_message_iter_get_arg_type(&iter_array) != DBUS_TYPE_INVALID;
+			dbus_message_iter_next(&iter_array)) {
+		if (dbus_message_iter_get_arg_type(&iter_array) != DBUS_TYPE_STRING)
+			goto fail;
+		if (i < *length)
+			dbus_message_iter_get_basic(&iter_array, &strings[i]);
+		i++;
+	}
+
+	*length = i;
+	return TRUE;
+
+fail:
+	signature = dbus_message_iter_get_signature(iter);
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect signature: %s != as", signature);
+	dbus_free(signature);
+	return FALSE;
+}
+
 /**
  * Call the given function for each key/value pairs. */
 dbus_bool_t bluealsa_dbus_message_iter_dict(
@@ -492,11 +705,6 @@ dbus_bool_t bluealsa_dbus_message_iter_get_pcm(
 	dbus_message_iter_get_basic(iter, &path);
 	strncpy(pcm->pcm_path, path, sizeof(pcm->pcm_path) - 1);
 
-	if (strstr(path, "/a2dp") != NULL)
-		pcm->flags |= BA_PCM_FLAG_PROFILE_A2DP;
-	if (strstr(path, "/sco") != NULL)
-		pcm->flags |= BA_PCM_FLAG_PROFILE_SCO;
-
 	if (!dbus_message_iter_next(iter))
 		goto fail;
 
@@ -534,24 +742,36 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_pcm_props_cb(const char *key,
 		strncpy(pcm->device_path, tmp, sizeof(pcm->device_path) - 1);
 		path2ba(tmp, &pcm->addr);
 	}
-	else if (strcmp(key, "Modes") == 0) {
-		if (type != DBUS_TYPE_ARRAY &&
-				dbus_message_iter_get_element_type(variant) != DBUS_TYPE_STRING) {
-			dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
-					"Incorrect variant for '%s': %c%c != as", key,
-					type, dbus_message_iter_get_element_type(variant));
-			return FALSE;
-		}
-		DBusMessageIter iter_props_modes;
-		for (dbus_message_iter_recurse(variant, &iter_props_modes);
-				dbus_message_iter_get_arg_type(&iter_props_modes) != DBUS_TYPE_INVALID;
-				dbus_message_iter_next(&iter_props_modes)) {
-			dbus_message_iter_get_basic(&iter_props_modes, &tmp);
-			if (strcmp(tmp, "source") == 0)
-				pcm->flags |= BA_PCM_FLAG_SOURCE;
-			else if (strcmp(tmp, "sink") == 0)
-				pcm->flags |= BA_PCM_FLAG_SINK;
-		}
+	else if (strcmp(key, "Sequence") == 0) {
+		if (type != (type_expected = DBUS_TYPE_UINT32))
+			goto fail;
+		dbus_message_iter_get_basic(variant, &pcm->sequence);
+	}
+	else if (strcmp(key, "Transport") == 0) {
+		if (type != (type_expected = DBUS_TYPE_STRING))
+			goto fail;
+		dbus_message_iter_get_basic(variant, &tmp);
+		if (strstr(tmp, "A2DP-source") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_A2DP_SOURCE;
+		else if (strstr(tmp, "A2DP-sink") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_A2DP_SINK;
+		else if (strstr(tmp, "HFP-AG") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HFP_AG;
+		else if (strstr(tmp, "HFP-HF") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HFP_HF;
+		else if (strstr(tmp, "HSP-AG") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HSP_AG;
+		else if (strstr(tmp, "HSP-HS") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HSP_HS;
+	}
+	else if (strcmp(key, "Mode") == 0) {
+		if (type != (type_expected = DBUS_TYPE_STRING))
+			goto fail;
+		dbus_message_iter_get_basic(variant, &tmp);
+		if (strcmp(tmp, "source") == 0)
+			pcm->mode = BA_PCM_MODE_SOURCE;
+		else if (strcmp(tmp, "sink") == 0)
+			pcm->mode = BA_PCM_MODE_SINK;
 	}
 	else if (strcmp(key, "Format") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
@@ -569,15 +789,21 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_pcm_props_cb(const char *key,
 		dbus_message_iter_get_basic(variant, &pcm->sampling);
 	}
 	else if (strcmp(key, "Codec") == 0) {
-		if (type != (type_expected = DBUS_TYPE_UINT16))
+		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->codec);
+		dbus_message_iter_get_basic(variant, &tmp);
+		strncpy(pcm->codec, tmp, sizeof(pcm->codec) - 1);
 	}
 	else if (strcmp(key, "Delay") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
 			goto fail;
 		dbus_message_iter_get_basic(variant, &pcm->delay);
 	}
+	else if (strcmp(key, "SoftVolume") == 0) {
+		if (type != (type_expected = DBUS_TYPE_BOOLEAN))
+			goto fail;
+		dbus_message_iter_get_basic(variant, &pcm->soft_volume);
+	}
 	else if (strcmp(key, "Volume") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
 			goto fail;
