@@ -1,6 +1,6 @@
 /*
  * test-io.c
- * Copyright (c) 2016-2019 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -18,26 +18,46 @@
 #include <check.h>
 
 #include "inc/sine.inc"
+#include "../src/a2dp.c"
+#include "../src/a2dp-audio.c"
 #include "../src/at.c"
+#include "../src/audio.c"
 #include "../src/ba-adapter.c"
 #include "../src/ba-device.c"
 #include "../src/ba-transport.c"
 #include "../src/bluealsa.c"
+#if ENABLE_APTX || ENABLE_APTX_HD
+# include "../src/codec-aptx.c"
+#endif
+#if ENABLE_MSBC
+# include "../src/codec-msbc.c"
+#endif
+#include "../src/codec-sbc.c"
+#include "../src/dbus.c"
+#include "../src/hci.c"
 #include "../src/io.c"
-#include "../src/msbc.c"
-#include "../src/rfcomm.c"
+#include "../src/sco.c"
 #include "../src/utils.c"
 #include "../src/shared/ffb.c"
 #include "../src/shared/log.c"
 #include "../src/shared/rt.c"
 
-int bluealsa_dbus_transport_register(struct ba_transport *t, GError **error) {
-	debug("%s: %p", __func__, (void *)t); (void)error;
-	return 0; }
-void bluealsa_dbus_transport_update(struct ba_transport *t, unsigned int mask) {
-	debug("%s: %p %#x", __func__, (void *)t, mask); }
-void bluealsa_dbus_transport_unregister(struct ba_transport *t) {
-	debug("%s: %p", __func__, (void *)t); }
+unsigned int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm, GError **error) {
+	debug("%s: %p", __func__, (void *)pcm); (void)error; return 0; }
+void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
+	debug("%s: %p %#x", __func__, (void *)pcm, mask); }
+void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm) {
+	debug("%s: %p", __func__, (void *)pcm); }
+struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
+	debug("%s: %p", __func__, (void *)sco); (void)fd; return NULL; }
+void ba_rfcomm_destroy(struct ba_rfcomm *r) {
+	debug("%s: %p", __func__, (void *)r); }
+int ba_rfcomm_send_signal(struct ba_rfcomm *r, enum ba_rfcomm_signal sig) {
+	debug("%s: %p: %#x", __func__, (void *)r, sig); return 0; }
+bool bluez_a2dp_set_configuration(const char *current_dbus_sep_path,
+		const struct a2dp_sep *sep, GError **error) {
+	debug("%s: %s", __func__, current_dbus_sep_path); (void)sep;
+	(void)error; return false; }
 
 static const a2dp_sbc_t config_sbc_44100_stereo = {
 	.frequency = SBC_SAMPLING_FREQ_44100,
@@ -58,7 +78,7 @@ static const a2dp_mpeg_t config_mp3_44100_stereo = {
 };
 
 static const a2dp_aac_t config_aac_44100_stereo = {
-	.object_type = AAC_OBJECT_TYPE_MPEG4_AAC_LC,
+	.object_type = AAC_OBJECT_TYPE_MPEG2_AAC_LC,
 	AAC_INIT_FREQUENCY(AAC_SAMPLING_FREQ_44100)
 	.channels = AAC_CHANNELS_2,
 	.vbr = 1,
@@ -90,50 +110,86 @@ static const char *input_pcm_file = NULL;
 static unsigned int aging_duration = 0;
 static bool dump_data = false;
 
+static void *write_test_pcm_async(void *userdata) {
+
+	int fd_out = (uintptr_t)userdata;
+	int fd_in;
+
+	ck_assert_int_ne(fd_in = open(input_pcm_file, O_RDONLY), -1);
+
+	struct pollfd pfds[] = {{ fd_out, POLLOUT, 0 }};
+	ssize_t len;
+
+	do {
+		ck_assert_int_ne(poll(pfds, ARRAYSIZE(pfds), -1), -1);
+		ck_assert_int_ne(len = sendfile(fd_out, fd_in, NULL, 0x7FFFF000), -1);
+	} while (len > 0);
+
+	close(fd_in);
+	return NULL;
+}
+
 /**
  * Write test PCM signal to the file descriptor. */
-static void write_test_pcm(int fd, int channels) {
+static void write_test_pcm(int fd, int channels, size_t frames) {
 
-	static int16_t sample_pcm_mono[5 * 1024];
-	static int16_t sample_pcm_stereo[10 * 1024];
+	static int16_t pcm_mono_buffer[1024];
+	static const size_t pcm_mono_samples = ARRAYSIZE(pcm_mono_buffer);
+	static int16_t pcm_stereo_buffer[2 * 1024];
+	static const size_t pcm_stereo_samples = ARRAYSIZE(pcm_stereo_buffer);
 	static bool initialized = false;
+	size_t samples = channels * frames;
+	pthread_t thread;
 	FILE *f;
-	int ffd;
 
 	if (input_pcm_file != NULL) {
-		ck_assert_int_ne(ffd = open(input_pcm_file, O_RDONLY), -1);
-		ck_assert_int_ne(sendfile(fd, ffd, NULL, 0x7FFFF000), -1);
-		close(ffd);
+		pthread_create(&thread, NULL, write_test_pcm_async, (void *)(uintptr_t)fd);
+		pthread_detach(thread);
 		return;
 	}
 
 	if (!initialized) {
 
 		/* initialize build-in sample PCM test signals */
-		snd_pcm_sine_s16le(sample_pcm_mono, ARRAYSIZE(sample_pcm_mono), 1, 0, 1.0 / 128);
-		snd_pcm_sine_s16le(sample_pcm_stereo, ARRAYSIZE(sample_pcm_stereo), 2, 0, 1.0 / 128);
+		snd_pcm_sine_s16le(pcm_mono_buffer, pcm_mono_samples, 1, 0, 1.0 / 128);
+		snd_pcm_sine_s16le(pcm_stereo_buffer, pcm_stereo_samples, 2, 0, 1.0 / 128);
 		initialized = true;
 
 		if (dump_data) {
 			ck_assert_ptr_ne(f = fopen("sample-mono.pcm", "w"), NULL);
-			fwrite(sample_pcm_mono, 1, sizeof(sample_pcm_mono), f);
+			fwrite(pcm_mono_buffer, 2, pcm_mono_samples, f);
 			fclose(f);
 			ck_assert_ptr_ne(f = fopen("sample-stereo.pcm", "w"), NULL);
-			fwrite(sample_pcm_stereo, 1, sizeof(sample_pcm_stereo), f);
+			fwrite(pcm_stereo_buffer, 2, pcm_stereo_samples, f);
 			fclose(f);
 		}
 
 	}
 
+	int16_t *pcm_buffer = NULL;
+	size_t pcm_samples = 0;
+
 	switch (channels) {
 	case 1:
-		ck_assert_int_eq(write(fd, sample_pcm_mono, sizeof(sample_pcm_mono)), sizeof(sample_pcm_mono));
+		pcm_buffer = pcm_mono_buffer;
+		pcm_samples = pcm_mono_samples;
 		break;
 	case 2:
-		ck_assert_int_eq(write(fd, sample_pcm_stereo, sizeof(sample_pcm_stereo)), sizeof(sample_pcm_stereo));
+		pcm_buffer = pcm_stereo_buffer;
+		pcm_samples = pcm_stereo_samples;
 		break;
 	}
 
+	ck_assert_ptr_ne(pcm_buffer, NULL);
+	ck_assert_int_ne(pcm_samples, 0);
+
+	while (samples > 0) {
+		size_t _samples = pcm_samples <= samples ? pcm_samples : samples;
+		size_t bytes = _samples * sizeof(*pcm_buffer);
+		ck_assert_int_eq(write(fd, pcm_buffer, bytes), bytes);
+		samples -= _samples;
+	}
+
 }
 
 struct bt_data {
@@ -160,6 +216,15 @@ static void bt_data_push(uint8_t *data, size_t len) {
 	bt_data_end = bt_data_end->next;
 }
 
+static void bt_data_write(int fd) {
+	struct bt_data *bt_data_head = &bt_data;
+	struct pollfd pfds[] = {{ fd, POLLOUT, 0 }};
+	for (; bt_data_head != bt_data_end; bt_data_head = bt_data_head->next) {
+		ck_assert_int_ne(poll(pfds, ARRAYSIZE(pfds), -1), -1);
+		ck_assert_int_eq(write(fd, bt_data_head->data, bt_data_head->len), bt_data_head->len);
+	}
+}
+
 /**
  * Helper function for timed thread join.
  *
@@ -193,13 +258,16 @@ static void test_a2dp_start_terminate_timer(unsigned int delay) {
 	pthread_detach(thread);
 }
 
-static void *test_io_thread_a2dp_dump_bt(void *arg) {
-	struct ba_transport *t = (struct ba_transport *)arg;
+static void *test_io_thread_a2dp_dump_bt(struct ba_transport_thread *th) {
 
+	struct ba_transport *t = th->t;
 	struct pollfd pfds[] = {{ t->bt_fd, POLLIN, 0 }};
 	uint8_t buffer[1024];
 	ssize_t len;
 
+	debug_transport_thread_loop(th, "START");
+	ba_transport_thread_ready(th);
+
 	while (poll(pfds, ARRAYSIZE(pfds), 500) > 0) {
 
 		if ((len = read(pfds[0].fd, buffer, sizeof(buffer))) == -1) {
@@ -221,9 +289,9 @@ static void *test_io_thread_a2dp_dump_bt(void *arg) {
 	return NULL;
 }
 
-static void *test_io_thread_a2dp_dump_pcm(void *arg) {
-	struct ba_transport *t = (struct ba_transport *)arg;
+static void *test_io_thread_a2dp_dump_pcm(struct ba_transport_thread *th) {
 
+	struct ba_transport *t = th->t;
 	struct pollfd pfds[] = {{ t->a2dp.pcm.fd, POLLIN, 0 }};
 	size_t decoded_samples_total = 0;
 	int16_t buffer[2048];
@@ -232,11 +300,13 @@ static void *test_io_thread_a2dp_dump_pcm(void *arg) {
 	FILE *f = NULL;
 
 	if (dump_data) {
-		sprintf(fname, "decoded-%s.pcm", bluetooth_a2dp_codec_to_string(t->type.codec));
+		sprintf(fname, "decoded-%s.pcm", ba_transport_codecs_a2dp_to_string(t->type.codec));
 		ck_assert_ptr_ne(f = fopen(fname, "w"), NULL);
 	}
 
-	debug("Starting PCM dump: %d", pfds[0].fd);
+	debug_transport_thread_loop(th, "START");
+	ba_transport_thread_ready(th);
+
 	while (poll(pfds, ARRAYSIZE(pfds), 500) > 0) {
 
 		if ((len = read(pfds[0].fd, buffer, sizeof(buffer))) == -1) {
@@ -244,14 +314,16 @@ static void *test_io_thread_a2dp_dump_pcm(void *arg) {
 			continue;
 		}
 
-		debug("Decoded samples: %zd", len / sizeof(int16_t));
-		decoded_samples_total += len / sizeof(int16_t);
+		size_t sample_size = BA_TRANSPORT_PCM_FORMAT_BYTES(t->a2dp.pcm.format);
+		debug("Decoded samples: %zd", len / sample_size);
+		decoded_samples_total += len / sample_size;
 
 		if (f != NULL)
 			fwrite(buffer, 1, len, f);
 	}
 
 	debug("Decoded samples total: %zd", decoded_samples_total);
+	ck_assert_int_gt(decoded_samples_total, 0);
 
 	if (f != NULL)
 		fclose(f);
@@ -262,10 +334,18 @@ static void *test_io_thread_a2dp_dump_pcm(void *arg) {
 	return NULL;
 }
 
+/**
+ * Adjust this variable in order to inject more PCM samples during A2DP test
+ * initialization. Remember to restore it to its original value afterwards. */
+static int test_a2dp_pcm_samples_boost = 1;
+
 /**
  * Drive PCM signal through A2DP source/sink loop. */
 static void test_a2dp(struct ba_transport *t1, struct ba_transport *t2,
-		void *(*enc)(void *), void *(*dec)(void *)) {
+		void *(*enc)(struct ba_transport_thread *), void *(*dec)(struct ba_transport_thread *)) {
+
+	const char *enc_name = enc == test_io_thread_a2dp_dump_pcm ? "dump-pcm" : "encode";
+	const char *dec_name = dec == test_io_thread_a2dp_dump_bt ? "dump-bt" : "decode";
 
 	int bt_fds[2];
 	int pcm_fds[2];
@@ -278,45 +358,38 @@ static void test_a2dp(struct ba_transport *t1, struct ba_transport *t2,
 
 	t1->type.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE;
 	t2->type.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	t1->state = TRANSPORT_ACTIVE;
-	t2->state = TRANSPORT_ACTIVE;
 	t1->bt_fd = bt_fds[1];
 	t2->bt_fd = bt_fds[0];
 	t1->a2dp.pcm.fd = pcm_fds[1];
 	t2->a2dp.pcm.fd = pcm_fds[0];
 
-	pthread_t thread1;
-	pthread_t thread2;
-
 	if (aging_duration)
 		test_a2dp_start_terminate_timer(aging_duration);
 
 	if (enc == test_io_thread_a2dp_dump_pcm) {
-		pthread_create(&thread2, NULL, dec, ba_transport_ref(t2));
-		struct bt_data *bt_data_head = &bt_data;
-		for (; bt_data_head != bt_data_end; bt_data_head = bt_data_head->next)
-			ck_assert_int_eq(write(bt_fds[1], bt_data_head->data, bt_data_head->len), bt_data_head->len);
-		pthread_create(&thread1, NULL, enc, ba_transport_ref(t1));
+		ck_assert_int_eq(ba_transport_thread_create(&t2->thread_dec, dec, dec_name), 0);
+		ck_assert_int_eq(ba_transport_thread_create(&t1->thread_enc, enc, enc_name), 0);
+		bt_data_write(bt_fds[1]);
 	}
 	else {
-		pthread_create(&thread1, NULL, enc, ba_transport_ref(t1));
-		write_test_pcm(pcm_fds[0], ba_transport_get_channels(t1));
-		pthread_create(&thread2, NULL, dec, ba_transport_ref(t2));
+		ck_assert_int_eq(ba_transport_thread_create(&t1->thread_enc, enc, enc_name), 0);
+		write_test_pcm(pcm_fds[0], t1->a2dp.pcm.channels, 4 * 1024 * test_a2dp_pcm_samples_boost);
+		ck_assert_int_eq(ba_transport_thread_create(&t2->thread_dec, dec, dec_name), 0);
 	}
 
 	pthread_mutex_lock(&test_a2dp_mutex);
 	pthread_cond_wait(&test_a2dp_terminate, &test_a2dp_mutex);
 	pthread_mutex_unlock(&test_a2dp_mutex);
 
-	ck_assert_int_eq(pthread_cancel(thread1), 0);
-	ck_assert_int_eq(pthread_cancel(thread2), 0);
+	ck_assert_int_eq(pthread_cancel(t1->thread_enc.id), 0);
+	ck_assert_int_eq(pthread_cancel(t2->thread_dec.id), 0);
 
-	ck_assert_int_eq(pthread_timedjoin(thread1, NULL, 1e6), 0);
-	ck_assert_int_eq(pthread_timedjoin(thread2, NULL, 1e6), 0);
+	ck_assert_int_eq(pthread_timedjoin(t1->thread_enc.id, NULL, 1e6), 0);
+	ck_assert_int_eq(pthread_timedjoin(t2->thread_dec.id, NULL, 1e6), 0);
 
 }
 
-static void test_sco(struct ba_transport *t, void *(*cb)(void *)) {
+static void test_sco(struct ba_transport *t, void *(*cb)(struct ba_transport_thread *)) {
 
 	int sco_fds[2];
 	int pcm_mic_fds[2];
@@ -325,15 +398,13 @@ static void test_sco(struct ba_transport *t, void *(*cb)(void *)) {
 	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET, 0, sco_fds), 0);
 	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_STREAM, 0, pcm_mic_fds), 0);
 	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_STREAM, 0, pcm_spk_fds), 0);
-	write_test_pcm(pcm_spk_fds[0], ba_transport_get_channels(t));
+	write_test_pcm(pcm_spk_fds[0], t->sco.spk_pcm.channels, 1024);
 
-	t->state = TRANSPORT_ACTIVE;
 	t->bt_fd = sco_fds[1];
 	t->sco.mic_pcm.fd = pcm_mic_fds[1];
 	t->sco.spk_pcm.fd = pcm_spk_fds[1];
 
-	pthread_t thread;
-	pthread_create(&thread, NULL, cb, ba_transport_ref(t));
+	ck_assert_int_eq(ba_transport_thread_create(&t->thread_enc, cb, "sco"), 0);
 
 	struct pollfd pfds[] = {
 		{ sco_fds[0], POLLIN, 0 },
@@ -357,16 +428,18 @@ static void test_sco(struct ba_transport *t, void *(*cb)(void *)) {
 
 		if (pfds[1].revents & POLLIN) {
 			ck_assert_int_gt(len = read(pcm_mic_fds[0], buffer, sizeof(buffer)), 0);
-			debug("Decoded samples: %zd", len / sizeof(int16_t));
-			decoded_samples_total += len / sizeof(int16_t);
+			size_t sample_size = BA_TRANSPORT_PCM_FORMAT_BYTES(t->sco.mic_pcm.format);
+			debug("Decoded samples: %zd", len / sample_size);
+			decoded_samples_total += len / sample_size;
 		}
 
 	}
 
 	debug("Decoded samples total: %zd", decoded_samples_total);
+	ck_assert_int_gt(decoded_samples_total, 0);
 
-	ck_assert_int_eq(pthread_cancel(thread), 0);
-	ck_assert_int_eq(pthread_timedjoin(thread, NULL, 1e6), 0);
+	ck_assert_int_eq(pthread_cancel(t->thread_enc.id), 0);
+	ck_assert_int_eq(pthread_timedjoin(t->thread_enc.id, NULL, 1e6), 0);
 
 	close(pcm_spk_fds[0]);
 	close(pcm_mic_fds[0]);
@@ -385,23 +458,25 @@ static int test_transport_release_bt_a2dp(struct ba_transport *t) {
 
 START_TEST(test_a2dp_sbc) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_SBC };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_SBC };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/sbc",
-			&config_sbc_44100_stereo, sizeof(config_sbc_44100_stereo));
+			&a2dp_codec_source_sbc, &config_sbc_44100_stereo);
 	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/sbc",
-			&config_sbc_44100_stereo, sizeof(config_sbc_44100_stereo));
+			&a2dp_codec_sink_sbc, &config_sbc_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
 	if (aging_duration) {
 		t1->mtu_write = t2->mtu_read = 153 * 3;
-		test_a2dp(t1, t2, io_thread_a2dp_source_sbc, io_thread_a2dp_sink_sbc);
+		test_a2dp(t1, t2, a2dp_source_sbc, a2dp_sink_sbc);
 	}
 	else {
 		t1->mtu_write = t2->mtu_read = 153 * 3;
-		test_a2dp(t1, t2, io_thread_a2dp_source_sbc, test_io_thread_a2dp_dump_bt);
-		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, io_thread_a2dp_sink_sbc);
+		test_a2dp(t1, t2, a2dp_source_sbc, test_io_thread_a2dp_dump_bt);
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_sbc);
 	}
 
 } END_TEST
@@ -409,48 +484,60 @@ START_TEST(test_a2dp_sbc) {
 #if ENABLE_MP3LAME
 START_TEST(test_a2dp_mp3) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_MPEG12 };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_MPEG12 };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/mp3",
-			&config_mp3_44100_stereo, sizeof(config_mp3_44100_stereo));
+			&a2dp_codec_source_mpeg, &config_mp3_44100_stereo);
 	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/mp3",
-			&config_mp3_44100_stereo, sizeof(config_mp3_44100_stereo));
+			&a2dp_codec_sink_mpeg, &config_mp3_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
+	int prev = test_a2dp_pcm_samples_boost;
+	/* We have to feed more samples while using MP3, because mpg123 library
+	 * does not output samples until at least two frames are decoded, where
+	 * one frame is 4094 samples. */
+	test_a2dp_pcm_samples_boost = 3;
+
 	if (aging_duration) {
 		t1->mtu_write = t2->mtu_read = 1024;
-		test_a2dp(t1, t2, io_thread_a2dp_source_mp3, io_thread_a2dp_sink_mpeg);
+		test_a2dp(t1, t2, a2dp_source_mp3, a2dp_sink_mpeg);
 	}
 	else {
 		t1->mtu_write = t2->mtu_read = 250;
-		test_a2dp(t1, t2, io_thread_a2dp_source_mp3, test_io_thread_a2dp_dump_bt);
-		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, io_thread_a2dp_sink_mpeg);
+		test_a2dp(t1, t2, a2dp_source_mp3, test_io_thread_a2dp_dump_bt);
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_mpeg);
 	}
 
+	test_a2dp_pcm_samples_boost = prev;
+
 } END_TEST
 #endif
 
 #if ENABLE_AAC
 START_TEST(test_a2dp_aac) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_MPEG24 };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_MPEG24 };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/aac",
-			&config_aac_44100_stereo, sizeof(config_aac_44100_stereo));
+			&a2dp_codec_source_aac, &config_aac_44100_stereo);
 	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/aac",
-			&config_aac_44100_stereo, sizeof(config_aac_44100_stereo));
+			&a2dp_codec_sink_aac, &config_aac_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
 	if (aging_duration) {
 		t1->mtu_write = t2->mtu_read = 450;
-		test_a2dp(t1, t2, io_thread_a2dp_source_aac, io_thread_a2dp_sink_aac);
+		test_a2dp(t1, t2, a2dp_source_aac, a2dp_sink_aac);
 	}
 	else {
 		t1->mtu_write = t2->mtu_read = 64;
-		test_a2dp(t1, t2, io_thread_a2dp_source_aac, test_io_thread_a2dp_dump_bt);
-		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, io_thread_a2dp_sink_aac);
+		test_a2dp(t1, t2, a2dp_source_aac, test_io_thread_a2dp_dump_bt);
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_aac);
 	}
 
 } END_TEST
@@ -459,20 +546,29 @@ START_TEST(test_a2dp_aac) {
 #if ENABLE_APTX
 START_TEST(test_a2dp_aptx) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_VENDOR_APTX };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_VENDOR_APTX };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/aptx",
-			&config_aptx_44100_stereo, sizeof(config_aptx_44100_stereo));
-	struct ba_transport *t2 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/aptx",
-			&config_aptx_44100_stereo, sizeof(config_aptx_44100_stereo));
+			&a2dp_codec_source_aptx, &config_aptx_44100_stereo);
+	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/aptx",
+			&a2dp_codec_sink_aptx, &config_aptx_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
 	if (aging_duration) {
+#if HAVE_APTX_DECODE
+		t1->mtu_write = t2->mtu_read = 400;
+		test_a2dp(t1, t2, a2dp_source_aptx, a2dp_sink_aptx);
+#endif
 	}
 	else {
 		t1->mtu_write = t2->mtu_read = 40;
-		test_a2dp(t1, t2, io_thread_a2dp_source_aptx, test_io_thread_a2dp_dump_bt);
+		test_a2dp(t1, t2, a2dp_source_aptx, test_io_thread_a2dp_dump_bt);
+#if HAVE_APTX_DECODE
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_aptx);
+#endif
 	};
 
 } END_TEST
@@ -481,20 +577,29 @@ START_TEST(test_a2dp_aptx) {
 #if ENABLE_APTX_HD
 START_TEST(test_a2dp_aptx_hd) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_VENDOR_APTX_HD };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_VENDOR_APTX_HD };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/aptxhd",
-			&config_aptx_hd_44100_stereo, sizeof(config_aptx_hd_44100_stereo));
-	struct ba_transport *t2 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/aptxhd",
-			&config_aptx_hd_44100_stereo, sizeof(config_aptx_hd_44100_stereo));
+			&a2dp_codec_source_aptx_hd, &config_aptx_hd_44100_stereo);
+	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/aptxhd",
+			&a2dp_codec_sink_aptx_hd, &config_aptx_hd_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
 	if (aging_duration) {
+#if HAVE_APTX_HD_DECODE
+		t1->mtu_write = t2->mtu_read = 600;
+		test_a2dp(t1, t2, a2dp_source_aptx_hd, a2dp_sink_aptx_hd);
+#endif
 	}
 	else {
 		t1->mtu_write = t2->mtu_read = 60;
-		test_a2dp(t1, t2, io_thread_a2dp_source_aptx_hd, test_io_thread_a2dp_dump_bt);
+		test_a2dp(t1, t2, a2dp_source_aptx_hd, test_io_thread_a2dp_dump_bt);
+#if HAVE_APTX_HD_DECODE
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_aptx_hd);
+#endif
 	};
 
 } END_TEST
@@ -503,20 +608,29 @@ START_TEST(test_a2dp_aptx_hd) {
 #if ENABLE_LDAC
 START_TEST(test_a2dp_ldac) {
 
-	struct ba_transport_type ttype = { .codec = A2DP_CODEC_VENDOR_LDAC };
+	struct ba_transport_type ttype = {
+		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+		.codec = A2DP_CODEC_VENDOR_LDAC };
 	struct ba_transport *t1 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/ldac",
-			&config_ldac_44100_stereo, sizeof(config_ldac_44100_stereo));
-	struct ba_transport *t2 = ba_transport_new_a2dp(device1, ttype, ":test", "/path/ldac",
-			&config_ldac_44100_stereo, sizeof(config_ldac_44100_stereo));
+			&a2dp_codec_source_ldac, &config_ldac_44100_stereo);
+	struct ba_transport *t2 = ba_transport_new_a2dp(device2, ttype, ":test", "/path/ldac",
+			&a2dp_codec_sink_ldac, &config_ldac_44100_stereo);
 
 	t1->acquire = t2->acquire = test_transport_acquire;
 	t1->release = t2->release = test_transport_release_bt_a2dp;
 
 	if (aging_duration) {
+#if HAVE_LDAC_DECODE
+		t1->mtu_write = t2->mtu_read = RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 990 + 6;
+		test_a2dp(t1, t2, a2dp_source_ldac, a2dp_sink_ldac);
+#endif
 	}
 	else {
-		t1->mtu_write = t2->mtu_read = RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 679;
-		test_a2dp(t1, t2, io_thread_a2dp_source_ldac, test_io_thread_a2dp_dump_bt);
+		t1->mtu_write = t2->mtu_read = RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 660 + 6;
+		test_a2dp(t1, t2, a2dp_source_ldac, test_io_thread_a2dp_dump_bt);
+#if HAVE_LDAC_DECODE
+		test_a2dp(t1, t2, test_io_thread_a2dp_dump_pcm, a2dp_sink_ldac);
+#endif
 	}
 
 } END_TEST
@@ -525,13 +639,13 @@ START_TEST(test_a2dp_ldac) {
 START_TEST(test_sco_cvsd) {
 
 	struct ba_transport_type ttype = { .profile = BA_TRANSPORT_PROFILE_HSP_AG };
-	struct ba_transport *t = ba_transport_new_sco(device1, ttype, ":test", "/path/sco/cvsd", NULL);
+	struct ba_transport *t = ba_transport_new_sco(device1, ttype, ":test", "/path/sco/cvsd", -1);
 
 	t->mtu_read = t->mtu_write = 48;
 	t->acquire = test_transport_acquire;
 
-	ba_transport_send_signal(t, TRANSPORT_PING);
-	test_sco(t, io_thread_sco);
+	ba_transport_thread_send_signal(&t->thread_enc, BA_TRANSPORT_SIGNAL_PING);
+	test_sco(t, sco_thread);
 
 } END_TEST
 
@@ -541,13 +655,13 @@ START_TEST(test_sco_msbc) {
 	struct ba_transport_type ttype = {
 		.profile = BA_TRANSPORT_PROFILE_HFP_AG,
 		.codec = HFP_CODEC_MSBC };
-	struct ba_transport *t = ba_transport_new_sco(device1, ttype, ":test", "/path/sco/msbc", NULL);
+	struct ba_transport *t = ba_transport_new_sco(device1, ttype, ":test", "/path/sco/msbc", -1);
 
 	t->mtu_read = t->mtu_write = 24;
 	t->acquire = test_transport_acquire;
 
-	ba_transport_send_signal(t, TRANSPORT_PING);
-	test_sco(t, io_thread_sco);
+	ba_transport_thread_send_signal(&t->thread_enc, BA_TRANSPORT_SIGNAL_PING);
+	test_sco(t, sco_thread);
 
 } END_TEST
 #endif
@@ -569,21 +683,23 @@ int main(int argc, char *argv[]) {
 		unsigned int flag;
 	} codecs[] = {
 #define TEST_CODEC_SBC  (1 << 0)
-		{ "SBC", TEST_CODEC_SBC },
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_SBC), TEST_CODEC_SBC },
 #define TEST_CODEC_MP3  (1 << 1)
-		{ "MP3", TEST_CODEC_MP3 },
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_MPEG12), TEST_CODEC_MP3 },
 #define TEST_CODEC_AAC  (1 << 2)
-		{ "AAC", TEST_CODEC_AAC },
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_MPEG24), TEST_CODEC_AAC },
 #define TEST_CODEC_APTX (1 << 3)
-		{ "APTX", TEST_CODEC_APTX },
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_VENDOR_APTX), TEST_CODEC_APTX },
 #define TEST_CODEC_APTX_HD (1 << 4)
-		{ "APTXHD", TEST_CODEC_APTX_HD },
-#define TEST_CODEC_LDAC (1 << 5)
-		{ "LDAC", TEST_CODEC_LDAC },
-#define TEST_CODEC_CVSD (1 << 6)
-		{ "CVSD", TEST_CODEC_CVSD },
-#define TEST_CODEC_MSBC (1 << 7)
-		{ "mSBC", TEST_CODEC_MSBC },
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_VENDOR_APTX_HD), TEST_CODEC_APTX_HD },
+#define TEST_CODEC_FASTSTREAM (1 << 5)
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_VENDOR_FASTSTREAM), TEST_CODEC_FASTSTREAM },
+#define TEST_CODEC_LDAC (1 << 6)
+		{ ba_transport_codecs_a2dp_to_string(A2DP_CODEC_VENDOR_LDAC), TEST_CODEC_LDAC },
+#define TEST_CODEC_CVSD (1 << 7)
+		{ ba_transport_codecs_hfp_to_string(HFP_CODEC_CVSD), TEST_CODEC_CVSD },
+#define TEST_CODEC_MSBC (1 << 8)
+		{ ba_transport_codecs_hfp_to_string(HFP_CODEC_MSBC), TEST_CODEC_MSBC },
 	};
 
 	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
@@ -626,7 +742,8 @@ int main(int argc, char *argv[]) {
 	SRunner *sr = srunner_create(s);
 
 	suite_add_tcase(s, tc);
-	tcase_set_timeout(tc, aging_duration + 5);
+	tcase_set_timeout(tc, aging_duration +
+			(input_pcm_file != NULL ? 180 : 5));
 
 	if (enabled_codecs & TEST_CODEC_SBC)
 		tcase_add_test(tc, test_a2dp_sbc);
